"""
Business logic for Vulture's mock logging, jobs, and dags
"""

import logging
import re
import subprocess
import sys
from collections import namedtuple
from contextlib import contextmanager
from pathlib import Path
from typing import Dict, List, NamedTuple

import pendulum
from dag import DAG

from workspaces.schema import WorkflowEventType


class MockLogger:
    """
    Class representing a logger for Vulture that will simulate HTCondor logs
    """

    def __init__(self, log_path: Path, out: str, err: str):
        self.out = out
        self.err = err
        self.logger = logging.getLogger("vulture")

        self.logger.setLevel(logging.INFO)
        self.def_handler = logging.FileHandler(f"{log_path}", mode="a+")
        self.logger.addHandler(self.def_handler)

    @contextmanager
    def switch_logger_file(self, log_file: str):
        """
        Switches logger FileHandler to write to a new file; operates as a context manager

        :param log_file: File name of the log you want to switch to
        """
        new_handler = logging.FileHandler(log_file, mode="w+")
        self.logger.removeHandler(self.def_handler)
        self.logger.addHandler(new_handler)
        yield
        self.logger.removeHandler(new_handler)
        self.logger.addHandler(self.def_handler)

    def write_out(self, which_out: str, output: str):
        """
        Writes output to an auxiliary log file (any file that is not the main condor.log)

        :param which_out: Which log is to be written to (out/error)
        :param output: Contents of the file
        :raises ValueError: If which_out is not a supported log type
        """
        # hello, world
        if which_out == "out":
            with self.switch_logger_file(self.out):
                self.logger.info(output)
        elif which_out == "err":
            with self.switch_logger_file(self.err):
                self.logger.info(output)
        else:
            # If adding new logging option, add to supported list
            supported = ["out", "err"]
            raise ValueError(
                f"Output '{which_out}' not supported. Use [{'|'.join(supported)}]."
            )

    def log_event(self, event_type: WorkflowEventType, event_body: str = None):
        """
        Log a mock HTCondor event to the log file

        :param event_type: Type of the event
        :param event_body: Body of the event log (can be empty)
        """
        header = self.create_header(event_type)
        # Indent each line of the body
        if event_body:
            event_body_list = event_body.split("\n")
            if len(event_body_list) > 1:
                event_body = "\t".join(event_body_list)
            else:
                event_body = "\t" + event_body
        footer = "..."
        self.logger.info(
            msg="\n".join([header, event_body, footer])
        ) if event_body else self.logger.info("\n".join([header, footer]))

    @staticmethod
    def create_header(event_type: WorkflowEventType) -> str:
        """
        Create mock HTCondor event log header

        EXAMPLE:
        000 (3983.000.000) 08/26 11:06:06 Job submitted from host: <10.64.1.178:9618?addrs=10.64.1.178-9618&noUDP&sock=4050180_7e37_3>

        :param event_type: Type of event being logged
        :raises ValueError: If event_type is not yet supported
        :return: The header as a string
        """
        mock_host = "<0.0.0.0:7777>"
        event_descs = {
            WorkflowEventType.SUBMITTED: f"Job submitted from host: {mock_host}",
            WorkflowEventType.EXECUTING: f"Job executing on host: {mock_host}",
            WorkflowEventType.TERMINATED: "Job terminated.",
            WorkflowEventType.OTHER: "Unknown description.",
        }
        event_num = f"{event_type.value:03d}"
        # TODO: Add calculable job number
        job_num = f"({4242:04d}.000.000)"
        timestamp = f"{pendulum.now().format('MM/DD HH:mm:ss')}"
        if event_type in event_descs.keys():
            event_desc = f"{event_descs[event_type]}"
        else:
            # If this gets triggered, add event type to event_descs dict above
            raise ValueError(f"Event type {event_type} not supported.")
        return " ".join([event_num, job_num, timestamp, event_desc])


class MalformedJobFile(ValueError):
    pass


class Job:
    """
    Class representing a faked HTCondor job
    """

    def __init__(self, file: Path, write_log: bool = True):
        self.fields, self.command = self.parse(file)
        self.write_log = write_log
        if self.write_log:
            if self.fields.get("executable"):
                # Collect logging file names from job file if they exist; else give them sane defaults
                exec_name = self.fields["executable"]
                self.logger = MockLogger(
                    log_path=self.fields.get("log", "./condor.log"),
                    out=self.fields.get("output", f"{exec_name}.out"),
                    err=self.fields.get("error", f"{exec_name}.error"),
                )
                # Job is submitted, write an event to signify that
                self.logger.log_event(
                    WorkflowEventType.SUBMITTED,
                    f"Job node: {self.fields['executable']}",
                )
            else:
                raise MalformedJobFile(
                    f"Condor job file {file} malformed. No executable found in fields {self.fields}."
                )

    def parse(self, file: Path) -> NamedTuple:
        """
        Parse condor job file for relevant fields
        TODO: Make parsing more robust

        :param file: Path to job file
        :return: NamedTuple of job data
        """
        # Parse file looking for executable and arguments
        JobData = namedtuple("JobData", ["field_data", "command"])
        fields = []

        try:
            with open(file, "r") as f:
                # Split file into fields and command
                for line in f.readlines():
                    if re.search(r" = ", line):
                        fields.append(line.rstrip())
                    else:
                        command = line.rstrip()
        except IOError as e:
            print(f"Could not open/read '{file}'.")
            sys.exit(1)

        if len(fields) > 0:
            field_dict = self.parse_fields(fields)
        else:
            raise MalformedJobFile(
                f"Condor job file '{file}' is malformed. No fields found."
            )
        return JobData(field_dict, command)

    @staticmethod
    def parse_fields(fields: List[str]) -> Dict[str, str]:
        """
        Take in list of fields parsed from input file and return them as a dictionary

        :param fields: List of fields
        :return: Dictionary of fields
        """
        field_dict = {}
        for field in fields:
            key, value = field.split(" = ")
            field_dict[key] = value
        return field_dict

    def execute(self):
        """
        Execute given process with given arguments
        TODO: Buff return code processing
        """
        result_strings = {
            0: "Normal termination (return value 0)",
            # This is 255 instead of -1 because that's what subprocess returns instead of -1
            255: "Error in execution (return value -1)",
        }
        capture_output = False
        if self.write_log:
            self.logger.log_event(WorkflowEventType.EXECUTING)
            capture_output = True

        # Execute
        if self.fields.get("executable"):
            process = subprocess.run(
                [self.fields["executable"], self.fields["arguments"]],
                capture_output=capture_output,
                text=True,
            )
        else:
            raise MalformedJobFile(
                f"Condor job file malformed. No executable found in fields {self.fields}."
            )

        exec_results = f"{result_strings.get(process.returncode, f'Something weird happened (return value {process.returncode})')}"
        if self.write_log:
            self.logger.log_event(WorkflowEventType.TERMINATED, exec_results)
            self.logger.write_out("out", process.stdout)
            self.logger.write_out("err", process.stderr)


class CondorDAG:
    """
    Class representing a faked HTCondor directed acyclic graph; uses py-dag.DAG library (https://github.com/thieman/py-dag)
    """

    def __init__(self, file: Path):
        """
        :param file: Path to HTCondor-DAG-formatted input file (see https://htcondor.readthedocs.io/en/latest/users-manual/dagman-workflows.html
        """
        dag = self.parse(file)

    @staticmethod
    def parse(file) -> DAG:
        # Initialize empty directed acyclic graph
        dag = DAG()

        return dag

    def execute(self):
        raise NotImplementedError
