"""
Services defined by the Workspaces system, to be used by our APIs and client programs.
"""

import os
import stat
import subprocess
import threading
import time
from collections import namedtuple
from pathlib import Path
from queue import Empty, PriorityQueue, Queue
from tempfile import mkdtemp
from typing import Dict, List, NamedTuple, Optional, Union

import requests
from channels.amqp_helpers import Channel, WorkflowEventChannel
from pycapo import CapoConfig
from sqlalchemy.orm import Session

from workspaces.workflow_interfaces import WorkflowRequestIF

from .capability_interfaces import (
    CapabilityEngineIF,
    CapabilityExecutionIF,
    CapabilityInfoIF,
    CapabilityName,
    CapabilityQueueIF,
    CapabilityRequestIF,
    CapabilityServiceIF,
    ParameterIF,
)
from .helpers import (
    CapabilityEventType,
    CapabilityRequestState,
    CapabilitySequence,
    CapabilityStepType,
    ExecutionPriority,
    ExecutionState,
    FutureProduct,
    Parameter,
    WorkflowRequestState,
)
from .product_interfaces import FutureProductIF
from .schema import (
    AbstractFile,
    Base,
    Capability,
    CapabilityEvent,
    CapabilityExecution,
    CapabilityRequest,
    CapabilityVersion,
    Workflow,
    WorkflowEvent,
    WorkflowEventType,
    WorkflowRequest,
    WorkflowRequestFile,
)
from .workflow_interfaces import WorkflowIF, WorkflowInfoIF, WorkflowServiceIF


class CapabilityService(CapabilityServiceIF):
    """
    The capability service: clients access this to request capability runs
    """

    def __init__(
        self, capability_info: CapabilityInfoIF, workflow_service: WorkflowServiceIF
    ):
        self.execution_pool = []
        self.queues = {}
        self.capability_info = capability_info
        self.workflow_service = workflow_service
        self.listener = threading.Thread(target=self.listen_for_events, daemon=True)
        self.listener.start()

    def create_request(
        self,
        capability_name: str,
        parameters: List[Parameter] = None,
        products: List[FutureProduct] = None,
    ) -> CapabilityRequestIF:
        """
        Create a capability request for a given capability

        :param capability_name:
        :param parameters: Parameters for request
        :param products: Products the request will create upon success
        :return: Capability request
        """
        return self.capability_info.create_capability_request(
            capability_name, parameters, products
        )

    def run_capability(self, request: CapabilityRequest) -> CapabilityExecutionIF:
        """
        Run a capability with the requested settings

        :param request: Capability request
        :return: Execution of the request
        """
        execution = self.capability_info.create_execution(request)
        steps = CapabilitySequence.from_str(execution.steps)

        if steps[0].step_type == CapabilityStepType.PrepareAndRunWorkflow:
            # First step is to run workflow, move to capability queue
            # FIXME: Priority needs to be dynamic
            self.enqueue_execution(execution, ExecutionPriority.Default.value)
        else:
            self.execution_pool.append(execution)
        return execution

    def update_execution(self, event: WorkflowEvent):
        """
        Update capability execution given a received event
        TODO: Implement execution cancellation

        :param event: Incoming event
        """
        capability_event = self.workflow_to_capability_event(event)

        if not capability_event:
            # WorkflowEvent does not need to update the execution
            return

        execution = self.capability_info.lookup_execution(capability_event.execution_id)
        step_sequence = CapabilitySequence.from_str(execution.steps)
        current_step = step_sequence[execution.current_step]
        event_type = CapabilityEventType[capability_event.event_type]

        # Check to make sure event type is correct
        if current_step.step_type.value == event_type.value:
            # Check if previous step (that just completed) is run workflow step
            if (
                step_sequence[execution.current_step].step_type
                == CapabilityStepType.PrepareAndRunWorkflow.name
            ):
                # Return capability engine to available state
                self.queues[execution.capability.id].complete_execution(
                    capability_event.execution_id
                )

            if execution.current_step != len(step_sequence):
                # Execution is not on its last step
                execution.current_step += 1
            else:
                # Execution on its last step and received the proper event
                execution.state = ExecutionState.Complete.name
                self.capability_info.save_execution(execution)
                return

            # Check if upcoming step is run workflow step
            if (
                step_sequence[execution.current_step].step_type
                == CapabilityStepType.PrepareAndRunWorkflow
            ):
                # Update execution record's state
                execution.state = ExecutionState.ExecutingStep.name
                # Enqueue execution that is on a run workflow step
                # FIXME: Priority needs to be dynamic; perhaps a priority column in execution schema?
                self.enqueue_execution(execution, ExecutionPriority.Default.value)

            # Update execution record's state
            execution.state = ExecutionState.Ready.name
            self.capability_info.save_execution(execution)
        else:
            # TODO: Logging
            print(
                f"Mismatched event type {event_type} for execution {capability_event.execution_id}"
            )

    def enqueue_execution(
        self,
        execution: CapabilityExecutionIF,
        priority: int = ExecutionPriority.Default.value,
    ):
        """
        Move execution record that is ready to execute a workflow into the appropriate capability
        queue

        :param execution: Capability execution record
        :param priority: Priority of the requested execution
        """
        # Get correct queue or initialize one
        queue = self.queues.get(
            execution.capability.id,
            CapabilityQueue(
                self.capability_info,
                self.workflow_service,
                execution.capability.max_jobs,
            ),
        )
        queue.enqueue(execution, priority)
        self.queues[execution.capability.id] = queue

        # Remove execution record from pool
        if execution in self.execution_pool:
            self.execution_pool.remove(execution)

    def workflow_to_capability_event(self, event: WorkflowEvent) -> CapabilityEvent:
        """
        Convert a workflow event into a capability event

        :param event: Workflow event
        :return: Capability event translated from given workflow event
        """
        if event.type == WorkflowEventType.SUBMITTED.name:
            # Workflow submitted
            event_type = CapabilityEventType.WorkflowReady
        elif event.type == WorkflowEventType.TERMINATED.name and event.retval == 0:
            # Workflow completed successfully
            event_type = CapabilityEventType.WorkflowComplete
        else:
            # No corresponding capability event
            return None

        execution = self.capability_info.lookup_execution_by_workflow_request_id(
            event.workflow_request_id
        )

        return CapabilityEvent(
            event_type, execution.capability_request.id, execution.id
        )

    def listen_for_events(self):
        """
        Method that will run in a different thread that will be listening for capability events that it can use
        to update capability executions
        """
        thread_workflow_events = Channel(WorkflowEventChannel(), threaded=True)
        thread_workflow_events.listen(callback=self.update_execution)


class CapabilityInfo(CapabilityInfoIF):
    """
    Interface for data access object that can look up and record information about capabilities
    and capability requests. Accesses and modifies the database.
    """

    def __init__(self, session: Session):
        self.session = session

    def create_capability(
        self, name: CapabilityName, steps: CapabilitySequence, max_jobs: int
    ) -> Capability:
        """
        Create new capability and save it in the database

        :param name: Name of new capability
        :param steps: List of capability steps required to execute the capability
        :param max_jobs: Max allowed number of concurrent executions of this kind of capability
        :return: Created capability
        """
        capability = Capability(name=name, steps=str(steps), max_jobs=max_jobs)
        self.save_entity(capability)
        return capability

    def create_capability_request(
        self,
        capability_name: str,
        parameters: List[ParameterIF] = None,
        future_products: List[FutureProductIF] = None,
    ) -> CapabilityRequest:
        """
        Create new capability request and save it in the database

        :param capability_name: name of the requested capability
        :param parameters: List
        :param future_products:
        :return: Created CapabilityRequest
        """
        capability = self.lookup_capability(capability_name)
        request = CapabilityRequest(
            state=CapabilityRequestState.Ready.name,
            capability=capability,
            parameters=str(parameters),
            # a trick here is to ensure that we always have a first version, with the original parameters
            versions=[CapabilityVersion(version_number=1, parameters=str(parameters))],
        )
        self.save_entity(request)
        return request

    def create_execution(self, request: CapabilityRequest) -> CapabilityExecution:
        """
        Create new execution record for a request and save it in the database

        :param request: ID of the capability request
        :return: Created CapabilityExecution
        """
        # look up the most recent version
        most_recent_version = request.versions[-1]

        execution = CapabilityExecution(
            state=ExecutionState.Ready.name,
            version=most_recent_version,
            current_step=0,
            # ensure that we have a copy of the step sequence as it was when the execution started
            steps=request.capability.steps,
        )
        self.save_entity(execution)
        return execution

    def lookup_capability(self, capability_name: str) -> Capability:
        return self.session.query(Capability).filter_by(name=capability_name).first()

    def lookup_capability_request(self, request_id) -> CapabilityRequest:
        return self.session.query(CapabilityRequest).filter_by(id=request_id).first()

    def lookup_execution(self, execution_id: int) -> CapabilityExecution:
        return (
            self.session.query(CapabilityExecution).filter_by(id=execution_id).first()
        )

    def save_entity(self, entity: Base):
        """
        Save entity to database

        :param entity:
        """
        self.session.add(entity)
        self.session.flush()

    def lookup_execution_by_workflow_request_id(
        self, workflow_request_id: int
    ) -> CapabilityExecution:
        """
        Find execution record given a capability request

        :param workflow_request_id:
        :return: Capability execution record
        """
        return (
            self.session.query(CapabilityExecution)
            .filter_by(workflow_request_id=workflow_request_id)
            .one()
        )

    def requests_for_capability(self, capability_name: str) -> List[CapabilityRequest]:
        """
        Gets all capability requests for a given capability

        :param capability_name: Name of capability
        :return: List of associated requests
        """
        capability = self.lookup_capability(capability_name)
        return (
            self.session.query(CapabilityRequest)
            .filter_by(capability_id=capability.id)
            .all()
        )

    def save_execution(self, execution: CapabilityExecutionIF):
        self.session.add(execution)


class CapabilityEngine(CapabilityEngineIF):
    """
    Executes a prepare and run workflow step of a capability
    """

    def __init__(
        self,
        capability_info: CapabilityInfoIF,
        workflow_service: "WorkflowServiceIF",
    ):
        self.capability_info = capability_info
        self.workflow_service = workflow_service

    def execute(self, execution: CapabilityExecutionIF):
        """
        Communicate with workflow service and send a request to run a workflow with given settings

        :param execution:
        """
        step_sequence = CapabilitySequence.from_str(execution.steps)
        # Grab value of current step (workflow name)
        cur_step = step_sequence[int(execution.current_step)]
        cur_step.execute(self, execution)

    def submit_workflow_request(
        self, workflow_name: str, workflow_args: dict, files: List[AbstractFile]
    ):
        """
        Submit a workflow request to the workflow service

        :param workflow_name:
        :param workflow_args:
        :param files:
        """
        workflow_request = WorkflowRequest(
            workflow_name=workflow_name, argument=workflow_args
        )
        self.workflow_service.execute(workflow_request, files)


class CapabilityQueue(CapabilityQueueIF):
    """
    Organizes capability executions in a priority order and makes it possible to control
    the number of concurrent executions or pause execution altogether
    """

    EngineList = namedtuple("EngineList", ["available", "in_use"])

    def __init__(
        self,
        capability_info: CapabilityInfoIF,
        workflow_service: WorkflowServiceIF,
        max_concurrency: int,
    ):
        self.capability_info = capability_info
        self.queue = PriorityQueue()
        self.max_concurrency = max_concurrency
        self.engine_list = self.init_engines(
            max_concurrency, capability_info, workflow_service
        )
        self.queue_manager = threading.Thread(
            target=self.process_executions, daemon=True
        )
        self.queue_manager.start()
        self.paused = False

    @classmethod
    def init_engines(
        cls,
        num_engines: int,
        capability_info: CapabilityInfoIF,
        workflow_service: WorkflowServiceIF,
    ) -> NamedTuple:
        """
        Initialize available engine queue and list of in-use engines

        :return: EngineList
        """
        engine_list = cls.EngineList(Queue(), [])
        for _ in range(num_engines):
            engine_list.available.put(
                CapabilityEngine(capability_info, workflow_service)
            )

        return engine_list

    def process_executions(self):
        """
        Process execution at the front of the queue; I imagine this would be called repeatedly
        with a small delay
        """
        while True:
            if self.queue.empty() is False:
                execution = self.queue.get()[1]
                engine = self.get_available_engine()
                if engine:
                    execution = self.capability_info.lookup_execution(execution.id)
                    # Start engine
                    engine.execute(execution)
                    # Move engine to in-use list
                    self.engine_list.in_use.append(engine)
                else:
                    # FIXME: Logging
                    print("No available engines. Try again later.")
            time.sleep(0.5)

    def find_engine(self, execution_id: int) -> CapabilityEngine:
        """
        Find in-use engine executing execution with given ID

        :param execution_id: ID of requested execution
        :return: Corresponding engine
        """
        for engine in self.engine_list.in_use:
            if engine.execution_id == execution_id:
                # Found correct engine
                return engine

    def complete_execution(self, execution_id: int):
        """
        Remove execution from its engine and return engine to available queue

        :param execution_id: ID of requested execution
        """
        engine = self.find_engine(execution_id)
        self.engine_list.in_use.remove(engine)
        engine.execution_id = None
        engine.execution = None
        self.engine_list.available.put(engine)

    def get_available_engine(self) -> Optional[CapabilityEngine]:
        """
        Get an engine if one is available

        :return: Engine if one is available, else False
        """
        try:
            return self.engine_list.available.get_nowait()
        except Empty:
            return None

    def enqueue(self, request: CapabilityExecutionIF, priority: int):
        self.queue.put((priority, request))

    def dequeue(self) -> CapabilityExecutionIF:
        return self.queue.get()

    def pause(self):
        # TODO: Implement
        self.paused = True

    def unpause(self):
        # TODO: Implement
        self.paused = False

    def __len__(self):
        return self.queue.qsize()


class WorkflowServiceRESTClient(WorkflowServiceIF):
    def __init__(self):
        self.url = (
            CapoConfig()
            .settings("edu.nrao.archive.workspaces.WorkflowSettings")
            .serviceUrl
        )

    def execute(self, request: WorkflowRequestIF, files: List["AbstractFile"]):
        # step 1: create the request
        req = requests.post(
            f"{self.url}/workflows/{request.workflow_name}/requests/create?args={request.argument}"
        ).json()

        # step 2: if necessary, pass the files up for this request
        for file in files:
            requests.post(
                f"{self.url}/workflows/requests/{req['workflow_request_id']}/files/{file.filename}",
                body=file.content,
            )

        # step 3: execute the request
        requests.post(
            f"{self.url}/workflows/requests/{req['workflow_request_id']}/submit"
        )


class WorkflowService(WorkflowServiceIF):
    """
    Executes workflows; should be a freestanding service.
    """

    # The next few things that need to happen here are in response to
    # WorkflowEvents that we receive from wf_monitor as the workflow
    # execution evolves. So we have to set up listening at some point
    # in this class
    def __init__(self, info: "WorkflowInfoIF"):
        self.info = info

        # Start listening for events from the wf_monitor stream
        self.listener = threading.Thread(
            target=lambda: Channel(WorkflowEventChannel(), threaded=True).listen(
                self.on_workflow_event
            )
        )
        self.listener.start()

    def execute(self, request: WorkflowRequest, files: List[AbstractFile] = None):
        """
        Execute a workflow per the supplied parameters.

        :param request: The Workflow Request to execute
        :param files: additional files to execute with request
        :return:
        """
        # 1. look up workflow, returns workflow
        definition = self.info.lookup_workflow_definition(request.workflow_name)

        # 2. render templates to files, returns list of rendered files
        workflow_files = definition.render_templates(request.argument, files)
        for file in workflow_files:
            self.info.save_file(
                request=request, filename=file.filename, content=file.content
            )

        # 4. prepare files for condor execution
        if not request.results_dir:
            # Create temp results directory if the request doesn't already have one (from a previous execution attempt)
            request.results_dir = str(
                self._prepare_files_for_condor(workflow_files).absolute()
            )
            self.info.save_request(request)

        # 5. execute condor and retrieve log file
        log_file = self._execute_prepared(Path(request.results_dir))

        # 6. start reading the log file (background it)
        subprocess.run(
            ["wf_monitor", log_file, str(request.workflow_request_id)],
            cwd=request.results_dir,
        )

    @staticmethod
    def _prepare_files_for_condor(files: List[WorkflowRequestFile]) -> Path:
        """
        Place the files for Condor into a new temp directory and returns the directory.

        :param files:  a dictionary of filename -> content
        :return:       a Path
        """
        # 1. create a temporary directory
        # TODO: decide where to put temp directory and make capo property
        if not os.path.isdir(str(Path.home()) + "/workspaces_tmp"):
            os.mkdir(str(Path.home()) + "/workspaces_tmp")
        temp_folder = Path(mkdtemp(dir=str(Path.home()) + "/workspaces_tmp"))

        # 2. spool each of the temp files to it
        for file in files:
            (temp_folder / file.filename).write_bytes(file.content)

        # 3. make any scripts in there executable
        for file in temp_folder.glob("*.sh"):
            file.chmod(file.stat().st_mode | stat.S_IEXEC)

        # finished, return folder
        return temp_folder

    @staticmethod
    def _execute_prepared(folder: Path) -> Path:
        """
        Execute HTCondor using the named folder as the source of the files.

        :param folder:  the path to the folder to execute
        :return:        the path to the log file
        """
        print(f"executing on folder {folder}")

        # some file in here should end in .dag; that file is our dagman input
        # TODO: not currently using DAG files for vulture
        dagman = list(folder.glob("*.dag"))[0]
        print(f"dagman file {dagman} exists.")

        # need .condor file for vulture submission
        condor = list(folder.glob("*.condor"))[0]
        print(f"condor file {condor} exists.")

        # ensure the log file exists
        logfile = folder / "condor.log"
        print(f"log file {logfile} exists.")
        logfile.touch()

        # submit
        # print("submitting to condor...")
        # subprocess.run(["condor_submit_dag", str(dagman)], cwd=str(folder.absolute()))

        # vulture is a workaround for testing locally without submitting to condor
        print("submitting to vulture...")
        subprocess.run(
            ["vulture", "job", "execute", str(condor)], cwd=str(folder.absolute())
        )

        # return the logfile
        return logfile

    def on_workflow_event(self, event: WorkflowEvent):
        request = self.info.lookup_workflow_request(event.workflow_request_id)

        status: str
        # 2. update request record with new status
        if event.type == WorkflowEventType.EXECUTING:
            status = WorkflowRequestState.Running.name
        elif event.type == WorkflowEventType.TERMINATED:
            if event.retval == 0:
                status = WorkflowRequestState.Complete.name
            else:
                status = WorkflowRequestState.Failed.name
        else:
            status = WorkflowRequestState.Running.name

        print(
            f"Updating state on workflow request {request.workflow_request_id} to {status}..."
        )
        request.update_status(status)


class WorkflowInfo(WorkflowInfoIF):
    """
    Implements WorkflowInfo, backed by a relational database, using SQL Alchemy.
    """

    def __init__(self, session):
        self.session = session

    def lookup_workflow_definition(self, name: str) -> Workflow:
        return self.session.query(Workflow).get(name)

    def lookup_workflow_request(self, request_id: int) -> WorkflowRequest:
        return self.session.query(WorkflowRequest).get(request_id)

    def all_workflows(self) -> List[Workflow]:
        return self.session.query(Workflow).all()

    def all_workflow_requests(self) -> List[WorkflowRequest]:
        """
        Query for the list of all workflow requests

        :return: List of workflow requests
        """
        return self.session.query(WorkflowRequest).all()

    def create_workflow_request(
        self, workflow: Union[str, WorkflowIF], argument: Dict
    ) -> WorkflowRequest:
        """
        Create new workflow request and save to database

        :param workflow: workflow to run
        :param argument: workflow arguments
        :return: new WorkflowRequest
        """

        # if it's an object with a workflow_name, then that's what we use,
        # otherwise, the argument must be already be a string
        workflow_name = (
            workflow.workflow_name if hasattr(workflow, "workflow_name") else workflow
        )

        request = WorkflowRequest(
            workflow_name=workflow_name,
            argument=argument,
            state=WorkflowRequestState.Created.name,
            results_dir="",
        )
        self.save_request(request)
        print(f"Created Workflow Request {request}")
        return request

    def save_request(self, request: WorkflowRequest):
        """
        Save a given request and return an integer identifier for it

        :param request: request to save
        """
        self.session.add(request)
        self.session.flush()

    def save_file(
        self, request: WorkflowRequest, filename: str, content: bytes
    ) -> WorkflowRequestFile:
        """
        Save a given file for the specified request to the database

        :param request: request to update with new file
        :param filename: filename of new file
        :param content: contents of new file in bytes
        :return:
        """
        wrf = WorkflowRequestFile(
            workflow_request_id=request.workflow_request_id,
            filename=filename,
            content=content,
        )
        self.session.add(wrf)
        self.session.flush()
        return wrf
